# frozen_string_literal: true

require 'ruby_jard/console'

require 'ruby_jard/decorators/color_decorator'
require 'ruby_jard/decorators/path_decorator'
require 'ruby_jard/decorators/loc_decorator'
require 'ruby_jard/decorators/source_decorator'
require 'ruby_jard/decorators/inspection_decorator'

require 'ruby_jard/screens'
require 'ruby_jard/color_schemes'
require 'ruby_jard/layouts'

require 'ruby_jard/row'
require 'ruby_jard/column'
require 'ruby_jard/span'
require 'ruby_jard/row_renderer'
require 'ruby_jard/screen_renderer'
require 'ruby_jard/screen_adjuster'
require 'ruby_jard/box_drawer'
require 'ruby_jard/screen_drawer'

module RubyJard
  ##
  # This class acts as a coordinator, in which it combines the data and screen
  # layout template, triggers each screen to draw on the terminal.
  class ScreenManager
    class << self
      extend Forwardable

      def_delegators :instance, :draw_screens, :puts, :draw_error, :start, :stop, :started?, :updating?

      def instance
        @instance ||= new
      end
    end

    attr_reader :output, :output_storage

    def initialize(output: RubyJard::Console.output)
      @output = output
      @screens = {}
      @started = false
      @updating = false
    end

    def start
      return if started?

      # Load configurations
      RubyJard.config
      RubyJard::Console.clear_screen(@output)

      @started = true
    end

    def started?
      @started == true
    end

    def updating?
      @updating == true
    end

    def stop
      return unless started?

      @started = false

      RubyJard::Console.cooked!(@output)
      RubyJard::Console.enable_echo!(@output)
      RubyJard::Console.enable_cursor!(@output)
    end

    def draw_screens
      start unless started?
      @updating = true

      RubyJard::Console.clear_screen(@output)
      RubyJard::Console.disable_cursor!(@output)
      width, height = RubyJard::Console.screen_size(@output)

      @layouts = calculate_layouts(width, height)
      @screens = build_screens(@layouts)

      RubyJard::Console.move_to(@output, 0, 0)

      draw_box(@screens)
      @screens.each do |screen|
        RubyJard::ScreenDrawer.new(
          output: @output,
          screen: screen,
          color_scheme: pick_color_scheme
        ).draw
      end

      RubyJard::Console.move_to(@output, 0, total_screen_height(@layouts) + 1)
      RubyJard::Console.clear_screen_to_end(@output)
    rescue StandardError => e
      RubyJard::Console.clear_screen(@output)
      draw_error(e, height)
    ensure
      # You don't want to mess up previous user TTY no matter happens
      RubyJard::Console.cooked!(@output)
      RubyJard::Console.enable_echo!(@output)
      RubyJard::Console.enable_cursor!(@output)
      @updating = false
    end

    def scroll; end

    def click; end

    def draw_error(exception, height = 0)
      @output.print RubyJard::Decorators::ColorDecorator::CSI_RESET
      @output.puts '--- Error ---'
      @output.puts "Internal error from Jard. I'm sorry to mess up your debugging experience."
      @output.puts 'It would be great if you can submit an issue in https://github.com/nguyenquangminh0711/ruby_jard/issues'
      @output.puts ''
      @output.puts exception
      if height == 0
        @output.puts exception.backtrace
      else
        @output.puts exception.backtrace.first(10)
      end
      @output.puts '-------------'
      RubyJard.error(exception)
    end

    private

    def calculate_layouts(width, height)
      layout = RubyJard::LayoutPicker.new(width, height).pick
      RubyJard::LayoutCalculator.calculate(
        layout_template: layout,
        width: width, height: height,
        x: 0, y: 0
      )
    end

    def build_screens(layouts)
      screens = layouts.map do |layout|
        screen_class = fetch_screen(layout.template.screen)
        screen = screen_class.new(layout)
        screen.build
        render_screen(screen)
        screen
      end
      RubyJard::ScreenAdjuster.new(screens).adjust
      layouts.map do |layout|
        screen_class = fetch_screen(layout.template.screen)
        screen = screen_class.new(layout)
        screen.build
        render_screen(screen)
        screen
      end
    end

    def draw_box(screens)
      RubyJard::BoxDrawer.new(
        output: @output,
        screens: screens,
        color_scheme: pick_color_scheme
      ).draw
    end

    def render_screen(screen)
      RubyJard::ScreenRenderer.new(
        screen: screen,
        color_scheme: pick_color_scheme
      ).render
    end

    def draw_debug(_width, height)
      @output.print RubyJard::Decorators::ColorDecorator::CSI_RESET
      unless RubyJard.debug_info.empty?
        @output.puts '--- Debug ---'
        RubyJard.debug_info.first(height - 2).each do |line|
          @output.puts line
        end
        @output.puts '-------------'
      end
      RubyJard.clear_debug
    end

    def fetch_screen(name)
      RubyJard::Screens[name]
    end

    def total_screen_height(layouts)
      layouts.map { |layout| layout.y + layout.height }.max || 0
    end

    def pick_color_scheme
      color_scheme_class =
        RubyJard::ColorSchemes[RubyJard.config.color_scheme] ||
        RubyJard::ColorSchemes[RubyJard::Config::DEFAULT_COLOR_SCHEME]
      color_scheme_class.new
    end
  end
end
